-- Objects
ATTACH_NONE         = 0
ATTACH_HAND         = 1
ATTACH_HEAD         = 2
ATTACH_RELATIVE     = 3
ATTACH_PROJECTILE   = 4

define_custom_obj_fields({
    oSmashHitboxGlobalOwner = 'u32',
    oSmashHitboxSize        = 'f32',
    oSmashHitboxDamage      = 'f32',
    oSmashHitboxKnockback   = 'f32',
    oSmashHitboxAngle       = 'u32',
    oSmashHitboxAttachment  = 'u32',
    oSmashHitboxScaling     = 'f32',
    oSmashHitboxPosX        = 'f32',
    oSmashHitboxPosY        = 'f32',
    oSmashHitboxPosZ        = 'f32',
})

E_MODEL_HITBOX = smlua_model_util_get_id("hitbox_geo")

---Create a hitbox.
---@param m MarioState
---@param model ModelExtendedId
---@param sound integer
---@param size number
---@param damage number
---@param baseKnockback number
---@param angle number
---@param x number
---@param y number
---@param z number
---@param xVel number
---@param yVel number
---@param zVel number
---@param attachPoint number
---@param life number
---@param scaling number How much the knockback scales with damage.
function create_hitbox(m, model, sound, size, damage, baseKnockback, angle, x, y, z, xVel, yVel, zVel, attachPoint, life, scaling)
    if m.playerIndex ~= 0 then return end
    local np = gNetworkPlayers[0]
    local s = gPlayerSyncTable[0]
    spawn_sync_object(id_bhvHitbox, model, x, y, z,
        function(obj)
            obj.oSmashHitboxGlobalOwner = np.globalIndex
            obj.oSmashHitboxSize = size*s.scale
            obj.oSmashHitboxDamage = damage*s.scale
            obj.oSmashHitboxKnockback = baseKnockback
            obj.oSmashHitboxAngle = angle
            obj.oSmashHitboxAttachment = attachPoint
            obj.oSmashHitboxScaling = scaling
            obj.oSmashHitboxPosX = x
            obj.oSmashHitboxPosY = y
            obj.oSmashHitboxPosZ = z
            obj.oVelX = xVel
            obj.oVelY = yVel
            obj.oVelZ = zVel
            obj.oHealth = life
            obj.oFaceAngleYaw = m.faceAngle.y
            obj.oSoundStateID = sound
            obj_scale(obj, 0)
            --obj_set_model_extended(obj, E_MODEL_HITBOX)
        end)
end

function intersects_local_player(obj, radius, playerRadius)
    local m = gMarioStates[0]

    local mPos = { x = m.pos.x, y = m.pos.y+50, z = m.pos.z }
    local oPos = { x = obj.oPosX, y = obj.oPosY, z = obj.oPosZ}
    
    return vec3f_dist(oPos, mPos) < radius+playerRadius
end

function intersects_any_player(m, obj, radius, playerRadius)

    local mPos = { x = m.pos.x, y = m.pos.y+50, z = m.pos.z }
    local oPos = { x = obj.oPosX, y = obj.oPosY, z = obj.oPosZ}
    
    return vec3f_dist(oPos, mPos) < radius+playerRadius
end

function bhv_hitbox_init(obj)
    local ownerNp = network_player_from_global_index(obj.oSmashHitboxGlobalOwner)
    local ownerM = gMarioStates[ownerNp.localIndex]

    local s = gPlayerSyncTable[ownerM.playerIndex]
    
    -- Initialize
    obj.oPosX = ownerM.pos.x + obj.oSmashHitboxPosX*s.scale
    obj.oPosY = ownerM.pos.y + obj.oSmashHitboxPosY*s.scale
    obj.oPosZ = ownerM.pos.z + obj.oSmashHitboxPosZ*s.scale
    
    obj.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    network_init_object(obj, false, nil)
end

function bhv_hitbox_loop(obj)
    local ownerNp = network_player_from_global_index(obj.oSmashHitboxGlobalOwner)
    local ownerM = gMarioStates[ownerNp.localIndex]
    local m = gMarioStates[0]

    local s = gPlayerSyncTable[ownerM.playerIndex]

    -- Change how the hitbox follows the player.
    if (obj.oSmashHitboxAttachment == ATTACH_HEAD) then
        obj.oPosX = ownerM.marioBodyState.headPos.x
        obj.oPosY = ownerM.marioBodyState.headPos.y
        obj.oPosZ = ownerM.marioBodyState.headPos.z
    elseif (obj.oSmashHitboxAttachment == ATTACH_RELATIVE) then
        obj.oPosX = ownerM.pos.x + obj.oSmashHitboxPosX*s.scale
        obj.oPosY = ownerM.pos.y + obj.oSmashHitboxPosY*s.scale
        obj.oPosZ = ownerM.pos.z + obj.oSmashHitboxPosZ*s.scale
    elseif (obj.oSmashHitboxAttachment == ATTACH_PROJECTILE) then

        obj.oPosX = obj.oPosX + obj.oVelX*s.scale
        obj.oPosY = obj.oPosY + obj.oVelY*s.scale
        obj.oPosZ = obj.oPosZ + obj.oVelZ*s.scale

        if (obj_has_model_extended(obj, E_MODEL_RED_FLAME)) then
            obj_set_billboard(obj)
            obj_scale(obj, ((14-obj.oHealth)*0.5)*s.scale)
        elseif (obj_has_model_extended(obj, E_MODEL_WATER_BOMB)) then
            obj_set_billboard(obj)
            obj_scale(obj, 1)
        elseif (obj_has_model_extended(obj, E_MODEL_BLACK_BOBOMB)) then
            obj_scale(obj, 1.2)
            obj.oAnimations = gObjectAnimations.bobomb_seg8_anims_0802396C
            cur_obj_play_sound_2(SOUND_AIR_BOBOMB_LIT_FUSE)
            if (obj.oHealth == 1) then
                spawn_non_sync_object(id_bhvExplosion, E_MODEL_EXPLOSION, obj.oPosX, obj.oPosY, obj.oPosZ, function() end)
            end
        end
    elseif (obj_has_model_extended(obj, E_MODEL_TWEESTER)) then
        if ownerM.action ~= ACT_SLIDE_KICK and obj.oSmashHitboxAttachment ~= ATTACH_PROJECTILE then
            obj_mark_for_deletion(obj)
        end
        --obj_scale(obj, 0.05)
        --obj_set_face_angle(obj, 0x4000, get_network_area_timer(), 0)
    end
    local localS = gPlayerSyncTable[0]
    local hitboxRadius = (obj.oSmashHitboxSize*100);
    local playerHitbox = (gCharData[gPlayerSyncTable[0].character].size*100)*localS.scale;
    
    if m.action ~= ACT_EATEN then
        if (global_index_hurts_mario_state(obj.oSmashHitboxGlobalOwner, m) and intersects_local_player(obj, hitboxRadius, playerHitbox)) then
            if (m.action == ACT_AIR_KNOCKBACK and m.actionState == 0) then return end
            -- Calculate damage and knockback based on hitbox properties.
            calculate_damage(m, obj.oSmashHitboxDamage, obj.oSmashHitboxKnockback, obj.oSmashHitboxAngle, obj.oFaceAngleYaw, obj.oSmashHitboxScaling)
            -- Sound
            play_sound(obj.oSoundStateID, m.marioObj.header.gfx.cameraToObject)
            -- Delete hitbox locally if interacts with player.
            obj_mark_for_deletion(obj)
            --djui_chat_message_create("Local Hit")
            return
        end
        for i=0, MAX_PLAYERS-1 do
            local mi = gMarioStates[i]
            playerHitbox = (gCharData[gPlayerSyncTable[i].character].size*100)*s.scale;

            if (intersects_any_player(mi, obj, hitboxRadius, playerHitbox) and (obj.oSmashHitboxGlobalOwner ~= gNetworkPlayers[mi.playerIndex].globalIndex)) then
                -- Sound
                play_sound(obj.oSoundStateID, m.marioObj.header.gfx.cameraToObject)
                -- Delete hitbox locally if interacts with player.
                obj_mark_for_deletion(obj)
                --djui_chat_message_create("Player Hit "..i)
                return
            end
        end
    end

    obj.oHealth = obj.oHealth - 1

    if obj.oHealth <= 0 then
        -- Delete hitbox once it expires.
        obj_mark_for_deletion(obj)
    end
end

id_bhvHitbox = hook_behavior(nil, OBJ_LIST_GENACTOR, true, bhv_hitbox_init, bhv_hitbox_loop)

function bhv_bobomb_king_init(obj)
    obj.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    obj_scale(obj, obj.oSnufitScale)
    cur_obj_play_sound_2(SOUND_AIR_BOBOMB_LIT_FUSE)

    obj.oAnimations = gObjectAnimations.bobomb_seg8_anims_0802396C
    cur_obj_init_animation(1)
    network_init_object(obj, false, nil)
    obj.oGravity = -4
    obj.oHealth = 25
end

function bhv_bobomb_king_loop(obj)
    cur_obj_update_floor_height()
    cur_obj_move_using_vel_and_gravity()
    if obj.oPosY < obj.oFloorHeight then
        if obj.oAction == 0 then
            obj.oVelY = 28
            obj.oVelX = obj.oVelX * 0.75
            obj.oVelZ = obj.oVelZ * 0.75
            obj.oAction = 1
        else obj.oVelY = 0 obj.oAction = 2 end
        obj.oPosY = obj.oFloorHeight

    end
    if obj.oAction == 2 then
        obj.oVelX = obj.oVelX * 0.75
        obj.oVelZ = obj.oVelZ * 0.75
    end
    obj.oHealth = obj.oHealth - 1

    if obj.oHealth < 0 then
        spawn_non_sync_object(id_bhvExplosion, E_MODEL_EXPLOSION, obj.oPosX, obj.oPosY, obj.oPosZ, function(o) obj_scale(o, obj.oSnufitScale) end)
        local m = gMarioStates[0]
    
        local localS = gPlayerSyncTable[0]
        local hitboxRadius = (300*obj.oSnufitScale);
        local playerHitbox = (gCharData[gPlayerSyncTable[0].character].size*100)*localS.scale;
        if (intersects_local_player(obj, hitboxRadius, playerHitbox)) then
            local yaw = obj_angle_to_object(obj, m.marioObj)
            -- Calculate damage and knockback.
            calculate_damage(m, 16*obj.oSnufitScale, 48, 24, yaw, 0.4)
            -- Sound
            play_sound(obj.oSoundStateID, m.marioObj.header.gfx.cameraToObject)
            -- Delete hitbox locally if interacts with player.
        end
        
        cur_obj_play_sound_1(SOUND_GENERAL2_BOBOMB_EXPLOSION)

        obj_mark_for_deletion(obj)
    end

    if obj.oHealth % 8 == 0 then obj_set_model_extended(obj, E_MODEL_BOBOMB_BUDDY) else obj_set_model_extended(obj, E_MODEL_BLACK_BOBOMB) end
end

id_bhvBobombKing = hook_behavior(nil, OBJ_LIST_GENACTOR, true, bhv_bobomb_king_init, bhv_bobomb_king_loop)

function bhv_hurtbox_init(obj)
    obj.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
end

function bhv_hurtbox_loop(obj)
    local ownerNp = network_player_from_global_index(obj.oSmashHitboxGlobalOwner)
    local ownerM = gMarioStates[ownerNp.localIndex]

    obj.oPosX = ownerM.pos.x
    obj.oPosY = ownerM.pos.y+50
    obj.oPosZ = ownerM.pos.z
    
    local playerHitbox = gCharData[CT_MARIO].size*100;
    obj_scale(obj, gCharData[CT_MARIO].size)

end

---@diagnostic disable-next-line: missing-parameter
id_bhvHurtbox = hook_behavior(nil, OBJ_LIST_GENACTOR, true, bhv_hurtbox_init, bhv_hurtbox_loop)

-- define ball's custom fields
define_custom_obj_fields({
    oGlobalOwner = 'u32',
})

function bhv_kirby_cap_init(obj)
    obj.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
end

function bhv_kirby_cap_loop(obj)
    local ownerNp = network_player_from_global_index(obj.oSmashHitboxGlobalOwner)
    local ownerM = gMarioStates[ownerNp.localIndex]

    obj.oPosX = ownerM.marioBodyState.headPos.x
    obj.oPosY = ownerM.marioBodyState.headPos.y
    obj.oPosZ = ownerM.marioBodyState.headPos.z

    obj.oFaceAnglePitch = ownerM.marioBodyState.headAngle.x + ownerM.marioObj.header.gfx.angle.x
    obj.oFaceAngleYaw = ownerM.marioBodyState.headAngle.y + ownerM.marioObj.header.gfx.angle.y
    obj.oFaceAngleRoll = ownerM.marioBodyState.headAngle.z + ownerM.marioObj.header.gfx.angle.z


end

---@diagnostic disable-next-line: missing-parameter
id_bhvKirbyCap = hook_behavior(nil, OBJ_LIST_GENACTOR, true, bhv_kirby_cap_init, bhv_kirby_cap_loop)

function bhv_bowser_bomb_init(obj)
    obj.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    network_init_object(obj, true, nil)
    --obj_mark_for_deletion(obj)
end

function bhv_bowser_bomb_loop(obj)
    local m = gMarioStates[0]
    
    local hitboxRadius = 250;
    local playerHitbox = gCharData[gPlayerSyncTable[0].character].size*100;
    if (intersects_local_player(obj, hitboxRadius, playerHitbox)) then
        local yaw = obj_angle_to_object(obj, m.marioObj)
        -- Calculate damage and knockback.
        calculate_damage(m, 24, 48, 38, yaw, 0.8)
        cur_obj_play_sound_1(SOUND_GENERAL_BOWSER_BOMB_EXPLOSION)
        -- Sound
        play_sound(obj.oSoundStateID, m.marioObj.header.gfx.cameraToObject)
        -- Delete hitbox locally if interacts with player.
        obj_mark_for_deletion(obj)
        network_send_object(obj, true)
        return
    end

end

---@diagnostic disable-next-line: missing-parameter
id_bhvCustomBowserBomb = hook_behavior(id_bhvBowserBomb, OBJ_LIST_GENACTOR, true, bhv_bowser_bomb_init, bhv_bowser_bomb_loop)

function bhv_menu_background_init(obj)
    obj.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
end

function bhv_menu_background_loop(obj)
    obj.oFaceAnglePitch = obj.oFaceAnglePitch + 0x50
end

---@diagnostic disable-next-line: missing-parameter
id_bhvMenuBackground = hook_behavior(nil, OBJ_LIST_GENACTOR, true, bhv_menu_background_init, bhv_menu_background_loop)