-- Utils

texStockRing = get_texture_info("hud-stock-ring")
texStockToadBase = get_texture_info("hud-stock-toad-base")

function lerp(a,b,t) return a * (1-t) + b * t end

-- Taken from Arena
function global_index_hurts_mario_state(globalIndex, m)
    if globalIndex == gNetworkPlayers[m.playerIndex].globalIndex then
        return false
    end

    local npAttacker = network_player_from_global_index(globalIndex)
    if npAttacker == nil then
        return false
    end
    local sAttacker = gPlayerSyncTable[npAttacker.localIndex]
    local sVictim = gPlayerSyncTable[m.playerIndex]

    if sAttacker.team == 0 or sVictim.team == 0 then
        return true
    end

    return sAttacker.team ~= sVictim.team
end

function djui_hud_print_text_outline(message, x, y, baseCol, outlineCol, scale)

    djui_hud_set_color((outlineCol & 0xFF0000) >> 16, (outlineCol & 0x00FF00) >> 8, (outlineCol & 0x0000FF) >> 0, 255)
    djui_hud_print_text(message, x, y-1, scale)
    djui_hud_print_text(message, x-1, y, scale)
    djui_hud_print_text(message, x+1, y, scale)
    djui_hud_print_text(message, x, y+1, scale)

    djui_hud_set_color((baseCol & 0xFF0000) >> 16, (baseCol & 0x00FF00) >> 8, (baseCol & 0x0000FF), 255)
    djui_hud_print_text(message, x, y, scale)
end

function calculate_damage(m, dmg, baseKnockback, pitch, yaw, scaling)
    local angleToSM64 = (0x8000/180)*pitch
    local e = gMarioStateExtras[m.playerIndex]
    local weight = gCharData[gPlayerSyncTable[0].character].weight

    local perc = gPlayerSyncTable[0].hp

    local isBoss = gGlobalSyncTable.gameMode == GAME_MODE_BOSS and gGlobalSyncTable.bossPlayer == gNetworkPlayers[0].globalIndex
    
    local armour = 0
    
    if (gGlobalSyncTable.gameMode == GAME_MODE_STAMINA or isBoss) then
        gPlayerSyncTable[0].hp = gPlayerSyncTable[0].hp - dmg
        perc = 0
    else
        gPlayerSyncTable[0].hp = gPlayerSyncTable[0].hp + dmg
    end
    
    if gNetworkPlayers[0].globalIndex == gGlobalSyncTable.bossPlayer and gGlobalSyncTable.gameMode == GAME_MODE_BOSS then
        armour = 65
    end

    local knockbackCalc = (((((perc*0.45)) + ((perc*dmg)*0.05)*weight*1.4)*scaling) + baseKnockback)-armour
    

    local endAction = ACT_GROUND_KNOCKBACK
    if (math.abs(baseKnockback) > 8 or m.action & ACT_FLAG_AIR ~= 0) then
        endAction = ACT_AIR_KNOCKBACK
        if (m.action & ACT_FLAG_AIR) ~= 0 and isBoss then
            armour = 10
        end
        knockbackCalc = (((((perc*0.35)) + ((perc*dmg)*0.05)*weight*1.4)*scaling) + math.abs(baseKnockback))-armour
    end

    if knockbackCalc <= 0 then
        return
    end

    set_mario_action(m, endAction, 0)
    m.faceAngle.y = yaw
    m.vel.x = (knockbackCalc*sins(yaw))*coss(angleToSM64)
    m.vel.y = knockbackCalc*sins(angleToSM64)
    m.vel.z = (knockbackCalc*coss(yaw))*coss(angleToSM64)
end

texFont = get_texture_info("hud-font")

function djui_hud_set_clamped_color(r, g, b, a)
    djui_hud_set_color(clamp(r, 0, 255), clamp(g, 0, 255), clamp(b, 0, 255), a)
end


function render_text(message, font, x, y, perRow, rowOffset, e)
    local xOffset = 0
    local imageSize = 256
    local charSize = imageSize/perRow
    local xHitOffset = 0
    local yHitOffset = 0
    

    for i=1, string.len(message) do
        local char = string.sub(message, i, i)
        if not tonumber(char) and not (char == "/") and not (char == "%") then
            local column = (string.byte(char)-97)%perRow
            local row = math.floor((string.byte(char)-97)/perRow)
            
            if char ~= " " then
                djui_hud_render_texture_tile(texFont, math.floor(x+i*14)+xOffset, math.floor(y), 1, 1, column*charSize, row*charSize, charSize, charSize)
            end
            if char == "i" then
                xOffset = xOffset - 8
            end
        else
            local column = (string.byte(char)-48)%perRow
            local row = math.floor((string.byte(char)-48)/perRow)+rowOffset

            if e ~= nil then
                local dir = (e.healthDirection+i*20)

                xHitOffset = math.floor(e.healthOffset*math.sin(dir))
                yHitOffset = math.floor(e.healthOffset*math.cos(dir))
            end
            
            if char == "%" then
                column = 2
                row = 1+rowOffset
            end

            if char == "/" then
                column = 3
                row = 1+rowOffset
            end

            djui_hud_render_texture_tile(texFont, math.floor(x+i*20)+xOffset+xHitOffset, math.floor(y)+yHitOffset+1, 1, 1, column*charSize, (row*charSize), charSize, charSize)
            if char == "1" then
                xOffset = xOffset - 4
            end
        end
    end
end

-- Thank PeachyPeach uwu
function relerp(x, a, b, c, d)
    return c + (d - c) * ((x - a) / (b - a))
end

function get_color(p)
    if (gGlobalSyncTable.gameMode == GAME_MODE_STAMINA) then
        return {
            r = 255,
            g = 255,
            b = 255,
        }
    end
    local colors = {
        [0] = { from =   0, to =  35, r0 = 0xFF, g0 = 0xFF, b0 = 0xFF, r1 = 0xFF, g1 = 0xFF, b1 = 0xAA }, -- 0% to 50%: white to yellow
        [1] = { from =  35, to = 70, r0 = 0xFF, g0 = 0xFF, b0 = 0xAA, r1 = 0xFF, g1 = 0x80, b1 = 0x40 }, -- 50% to 100%: yellow to orange
        [2] = { from = 70, to = 105, r0 = 0xFF, g0 = 0x80, b0 = 0x40, r1 = 0xDD, g1 = 0x10, b1 = 0x20 }, -- 100% to 150%: orange to red
        [3] = { from = 105, to = 140, r0 = 0xDD, g0 = 0x10, b0 = 0x20, r1 = 0x60, g1 = 0x00, b1 = 0x00 }, -- 150% to 200%: red to dark red
    }
    p = math.min(p, 140)
    local gradient = colors[math.min(3, math.floor(p / 35))]
    return {
        r = relerp(p, gradient.from, gradient.to, gradient.r0, gradient.r1),
        g = relerp(p, gradient.from, gradient.to, gradient.g0, gradient.g1),
        b = relerp(p, gradient.from, gradient.to, gradient.b0, gradient.b1),
    }
end

function convert_s16(num)
    local min = -32768
    local max = 32767
    while (num < min) do
        num = max + (num - min)
    end
    while (num > max) do
        num = min + (num - max)
    end
    return num
end

camYaw = 0
camYawSpeed = 0

camPitch = 0
camPitchSpeed = 0

camX = 0
camY = 200
camZ = 0

camZoomLevel = 0
camZoom = 1000

camForward = 200

targetState = 1

lTargetSpeed = 0.3

targetM = nil

lastMouseX = djui_hud_get_raw_mouse_x()*4
lastMouseY = djui_hud_get_raw_mouse_y()*4

function smash_camera()
    local camSignX = 1
    local camSignY = 1
    camera_config_enable_free_cam(false)
    
    
    if not camera_config_is_x_inverted() then camSignX = -1 end
    if not camera_config_is_y_inverted() then camSignY = -1 end
    local xSens = camera_config_get_x_sensitivity()*0.5
    local ySens = camera_config_get_y_sensitivity()*0.5
    camera_freeze()
    
    local s = gPlayerSyncTable[0]

    local camTargetY = 200
    local m = gMarioStates[0]

    if math.abs(m.controller.extStickX) > 64 or math.abs((djui_hud_get_raw_mouse_x()-lastMouseX)*4) > 400 then
        targetM = nil
    end

    local mPos = m.pos -- Local Position

    local grounded = m.floorHeight-m.pos.y <= 20
    
    local targetPitch = nil
    local pitchDiff = 0

    local currYaw = gLakituState.yaw
    local targetYaw = nil
    local yawDiff = 0

    if (m.controller.buttonDown & L_TRIG) ~= 0 then
        lTargetSpeed = 0.3
        targetM = nearest_mario_state_to_object(m.marioObj)
    end

    if (m.controller.buttonPressed & R_TRIG) ~= 0 then
        camZoomLevel = camZoomLevel + 1
    end

    camZoomLevel = (camZoomLevel % 3)
    
    camZoom = lerp(camZoom, 1000+(camZoomLevel*400)*s.scale, 0.5)
    
    camForward = 300

    if (targetM ~= nil) then
        camForward = 0
        targetYaw = obj_angle_to_object(gMarioStates[0].marioObj, targetM.marioObj)+0x8000
        targetPitch = obj_pitch_to_object(m.marioObj, targetM.marioObj)
        yawDiff = convert_s16(targetYaw-currYaw)*lTargetSpeed
        pitchDiff = (convert_s16(targetPitch-camPitch)*lTargetSpeed)*0.5
    end

    lTargetSpeed = lerp(lTargetSpeed, 0.1, 0.1)

    local pan = {x = 0, y = 0, z = 0}
    pan.y = sins(0xC00) * 400
    rotate_in_xz(pan, pan, m.faceAngle.y)
    rotate_in_xz(pan, pan, -gLakituState.yaw)

    gLakituState.focHSpeed = 0
    gLakituState.focVSpeed = 0
    
    camX = lerp(camX, sins(m.faceAngle.y)*camForward, 0.05)
    camZ = lerp(camZ, coss(m.faceAngle.y)*camForward, 0.05)

    focusPosX = mPos.x
    focusPosZ = mPos.z

    vec3f_set(gLakituState.focus, focusPosX+camX, m.pos.y+camY, focusPosZ+camZ)
    gLakituState.focusDistance = 200

    local camYaw = m.controller.extStickX*xSens

    if camera_config_is_mouse_look_enabled() then
        camYaw = (djui_hud_get_raw_mouse_x()-lastMouseX)*4
    end

    camYawSpeed = approach_f32(camYawSpeed, camYaw, 100, 100)
    gLakituState.yaw = gLakituState.yaw + camYawSpeed*camSignX + yawDiff

    local t = 0

    if math.abs(m.controller.extStickY) > 70 then
        t = -m.controller.extStickY*ySens
    end
    if camera_config_is_mouse_look_enabled() then
        t = (djui_hud_get_raw_mouse_y()-lastMouseY)*4--*ySens
    end

    lastMouseX = djui_hud_get_raw_mouse_x()*4
    lastMouseY = djui_hud_get_raw_mouse_y()*4

    camPitchSpeed = approach_f32(camPitchSpeed, t, 100, 100)
    camPitch = clamp(camPitch + camPitchSpeed*camSignY, -0x3500, 0x3000) + pitchDiff
    if targetM ~= nil then
        camPitch = clamp(camPitch, -0x3500, 0x1000)
    end

    local mOffset = {x = focusPosX, y = mPos.y+200, z = focusPosZ}

    vec3f_set_dist_and_angle(mOffset, gLakituState.pos, camZoom, camPitch+0x700, gLakituState.yaw)
    if not (grounded) then
        camTargetY = 0
    end

    camY = lerp(camY, camTargetY, 0.5)

    gLakituState.oldPitch = camPitch

    if gLakituState.pos.y < m.floorHeight+1 then
        gLakituState.pos.y = m.floorHeight+1
    end

    gMarioStates[0].area.camera.yaw = gLakituState.yaw
end

function new_camera()
    local m = gMarioStates[0]
    local camSignX = 1
    local lS = gLakituState
    if not camera_config_is_x_inverted() then camSignX = -1 end
    camera_freeze()

    lS.yaw = lS.yaw + m.controller.extStickX*camSignX
end

-- From Football
function active_player(m)
    local np = gNetworkPlayers[m.playerIndex]
    if m.playerIndex == 0 then
        return true
    end
    if not np.connected then
        return false
    end
    return is_player_active(m)
end

playerPalettes = {

}

---Add new fighters.
---@param name string
---@param weight number
---@param size number
---@param topSpeed number
---@param acc number
---@param dcc number
---@param frc number
---@param charIcon TextureInfo
---@param model ModelExtendedId
---@param charType CharacterType
---@param charMarioUpdate function
---@param charBeforeMarioUpdate function
---@param stockRender function
---@param hudRender function
---@param jab function
---@param aerial function
---@param gP function
---@param gPA function
---@param gPB function
---@param slideKick function
---@param dash function
---@param pal table
function smash_add_character(name, weight, size, topSpeed, acc, dcc, frc, charIcon, model, charType, charMarioUpdate, charBeforeMarioUpdate, stockRender, hudRender, jab, aerial, gP, gPA, gPB, slideKick, dash, pal)
    table.insert(gCharData, {
        charId = #gCharData+1,
        name = name,
        weight = weight,
        size = size,
        topSpeed = topSpeed,
        acc = acc,
        dcc = dcc,
        frc = frc,
        charIcon = charIcon,
        model = model,
        charType = charType,
        charMarioUpdate = charMarioUpdate,
        charBeforeMarioUpdate = charBeforeMarioUpdate,
        stockRender = stockRender,
        hudRender = hudRender,
        jab = jab,
        aerial = aerial,
        groundPound = gP,
        groundPoundA = gPA,
        groundPoundB = gPB,
        slideKick = slideKick,
        dash = dash,
        pal = pal,
    })
end

function smash_add_stage(level, name, texture, radius, bottom, mus)
    table.insert(gSmashArenas, {
        level = level,
        name = name,
        texture = texture,
        radius = radius,
        bottom = bottom,
        mus = mus
    })
end